@@ -5,14 +5,13 @@
 
 #' An S4 class to store information
 #'
-#' @slot type One of two available types "dense" or "sparse". See description.
-#' @slot data A data.frame with covariates information, where colnames(data) matches model.list specification
-#' @slot locs a matrix with locs matching data
-#' @slot z A vector with response values
-#' @slot model.list A list specyfing a model for each aspect of the spatial structure.
-#' @slot info a list with information about the coco object
-#' @slot output an output from optimparallel output, including as well boundaries 
-#' information as another element of the list
+#' @slot type (\code{character}) One of two available types "dense" or "sparse". See description.
+#' @slot data (\code{data.frame}) A data.frame with covariates information, where colnames(data) matches model.list specification
+#' @slot locs (\code{numeric matrix}) a matrix with locs matching data
+#' @slot z (\code{numeric matrix}) A matrix of dimension n x p with response values
+#' @slot model.list (\code{list}) A list specifying a model for each aspect of the spatial structure.
+#' @slot info (\code{list}) a list with information about the coco object
+#' @slot output (\code{list}) if building an already fitted \code{coco} object (not the standard approach), then requires an output from Optimparallel output, including as well boundaries, etc.
 #' @author Federico Blasi
 #' 
 setClass("coco", slots = list(
@@ -29,26 +28,25 @@ setClass("coco", slots = list(
 # Methods ----------------------------------------------------------------------
 ###############################################################################-
 
-#' Plot Method for Coco Class
+#' Plot Method for coco objects
 #'
 #' This method plots objects of class \code{coco}.
-#'
-#' @param x An object of class \code{coco}.
-#' @param y Not used.
-#' @param ... Additional arguments passed to the plot function. when type "ellipse" , delta of nearest.dist must be specified.
-#' @param type The type of plot. NULL or "ellipse" for drawing ellipse of the convolution kernels.
-#' @param index For plotting local correlation plots.
-#' @param factr Factor rate for size of ellipses.
-#' @param plot.control Additional plot control parameters.
-#' @return A plot is created.
-#' @exportMethod plot
-#' @docType methods
 #' @rdname plot-methods
+#' @docType methods
 #' @aliases plot,coco-method
+#' @param x (\code{S4}) A fitted object of class \code{coco}.
+#' @param y Not used.
+#' @param type (\code{character}  or \code{NULL}) The type of plot. NULL or "ellipse" for drawing ellipse of the convolution kernels.
+#' @param index (\code{integer vector}) For plotting local correlation plots.
+#' @param factr (\code{numeric}) Factor rate for size of ellipses.
+#' @param ... Additional arguments passed to \link[fields]{quilt.plot}. 
+#' @return Several plots are created.
+#' @author Federico Blasi
+#' 
 setMethod("plot",
           signature(x = "coco", y = "missing"),
           definition =
-            function(x, y, ..., type = NULL, index = NULL, factr = 0.1, plot.control = NULL) {
+            function(x, y, type = NULL, index = NULL, factr = 0.1, ...) {
               
               if (length(x@output) == 0) {
                 stop("object has not yet been fitted.")
@@ -86,22 +84,22 @@ setMethod("plot",
                   fields::quilt.plot(x@locs, spat_effects$sd, main = "se", ...)
                   
                   #plot(x@locs[,1], x@locs[,2], pch=20, main = "se",
-                  #     col = tim.colors(64)[cut(tp_se, breaks = quantile(x = tp_se, probs = seq(0, 1, length.out = 64)), 
+                  #     col = coco.pallete(64)[cut(spat_effects$sd, breaks = quantile(x = spat_effects$sd, probs = seq(0, 1, length.out = 64)), 
                   #                              labels = FALSE, include.lowest = T)], xlab = colnames(x@locs)[1],ylab=colnames(x@locs)[2])
                   
                   fields::quilt.plot(x@locs, spat_effects$scale_x, main = "approx. eff. scale", ...)
                   
                   #plot(x@locs[,1], x@locs[,2], pch=20, main = "approx eff.scale",
-                  #     col = tim.colors(64)[cut(spat_effects$scale_x, breaks = quantile(x = spat_effects$scale_x, probs = seq(0, 1, length.out = 64)), 
+                  #     col = coco.pallete(64)[cut(spat_effects$scale_x, breaks = quantile(x = spat_effects$scale_x, probs = seq(0, 1, length.out = 64)), 
                   #                              labels = FALSE, include.lowest = T)], xlab = colnames(x@locs)[1],ylab=colnames(x@locs)[2])
                   
                   fields::quilt.plot(x@locs, spat_effects$aniso, main = "anisotropy", ...)
                   
                   #plot(x@locs[,1], x@locs[,2], pch=20, main = "ga",
-                  #     col = tim.colors(64)[cut(spat_effects$aniso, breaks = quantile(x = spat_effects$aniso, probs = seq(0, 1, length.out = 64)), 
+                  #     col = coco.pallete(64)[cut(spat_effects$aniso, breaks = quantile(x = spat_effects$aniso, probs = seq(0, 1, length.out = 64)), 
                   #                              labels = FALSE, include.lowest = T)], xlab = colnames(x@locs)[1],ylab=colnames(x@locs)[2])
                   
-                  fields::quilt.plot(x@locs, spat_effects$tilt, main = "tilt", zlim = c(-pi / 4 - 0.1, pi / 4 + 0.1), ...)
+                  fields::quilt.plot(x@locs, spat_effects$angle, main = "tilt", zlim = c(0, pi), ...)
                   fields::quilt.plot(x@locs, spat_effects$smooth, main = "smooth", ...)
                   fields::quilt.plot(x@locs, spat_effects$nugget, main = "nugget", ...)
                   
@@ -114,7 +112,7 @@ setMethod("plot",
                     
                     graphics::par(mfrow = c(1, 1))
                     
-                    plot(x@locs, col = fields::tim.colors(128)[cut(x@z, 128)], pch = 18, cex = 2, asp = 1)
+                    plot(x@locs, col = fields::tim.colors(128)[cut(x@z, 128)], pch = 20, cex = 1, asp = 1)
                     
                     number_x <- 10
                     range_x <- range(x@locs[, 1])
@@ -130,12 +128,16 @@ setMethod("plot",
                       for (jj in 1:number_x) {
                         center_locs <- c(vals_x[ii], range_y[1] + jj / number_x * (range_y[2] - range_y[1]))
                         
-                        sss <- spam::nearest.dist(x = matrix(center_locs, ncol = 2), y = x@locs, delta = ...) # fix delta to automatic
+                        sss <- spam::nearest.dist(x = matrix(center_locs, ncol = 2), y = x@locs, delta = diff(range(x@locs))/5)
+                        
+                        if(length(sss@entries) == 0){next}
                         
                         to_compute_sd <- sss@colindices[which.min(sss@entries)]
                         
                         .cocons.DrawEllipsoid(
-                          alpha_i = spat_effects$tilt[to_compute_sd], r = spat_effects$aniso[to_compute_sd], rho = spat_effects$scale_x[to_compute_sd],
+                          alpha_i = spat_effects$tilt[to_compute_sd], 
+                          r = spat_effects$aniso[to_compute_sd], 
+                          rho = spat_effects$scale_x[to_compute_sd],
                           loc = x@locs[to_compute_sd, ], factr = factr
                         )
                       }
@@ -143,12 +145,8 @@ setMethod("plot",
                   }
                 }
                 
-              }
-              
-              if (!is.null(type)) {
-                
-                # !!!!!!! is it ok && ??
-                if (x@type == "dense" && type == "correlations") {
+                if (!is.null(type)) {
+                if (type == "correlations") {
                   
                   tmp_info <- cocons::getDesignMatrix(model.list = x@model.list, data = x@data)
                   
@@ -170,7 +168,7 @@ setMethod("plot",
                   
                   for (ww in index) {
                     
-                    fields::quilt.plot(x@locs, tmp_cov[ww,], zlim = c(0, 1), main = 'global corr.')
+                    fields::quilt.plot(x@locs, tmp_cov[ww,], zlim = c(0, 1), main = paste0('global corr. at index ',ww))
                     graphics::points(x@locs[ww,1], x@locs[ww,2], pch = "X", bg='red', col = 'violet', cex = 2)
                     
                     local_var <- X_std$std.covs[ww, ] %*% theta_list$std.dev
@@ -199,7 +197,7 @@ setMethod("plot",
                       info = list("smooth.limits" = x@info$smooth.limits)
                     )
                     
-                    teteee <- getDesignMatrix(model.list = model.list, 
+                    DM_tmp <- getDesignMatrix(model.list = model.list, 
                                               data = local_object@data)
                     
                     here <- getModelLists(c(local_var, 
@@ -207,33 +205,25 @@ setMethod("plot",
                                             local_ga, 
                                             local_tt, 
                                             local_smtns),
-                                          par.pos = teteee$par.pos, type = "diff"
+                                          par.pos = DM_tmp$par.pos, type = "diff"
                     )
                     
                     tmp_cov_two <- stats::cov2cor(cocons::cov_rns(
                       theta = here, locs = x@locs,
-                      x_covariates = teteee$model.matrix,
+                      x_covariates = DM_tmp$model.matrix,
                       smooth_limits = x@info$smooth.limits
                     ))
                     
-                    fields::quilt.plot(x@locs, tmp_cov_two[ww, ], main = 'local corr.')
+                    fields::quilt.plot(x@locs, tmp_cov_two[ww, ], main = paste0('local corr. at index ',ww))
                     graphics::points(x@locs[ww,1], x@locs[ww,2], pch = "X", bg='red', col = 'violet', cex = 2)
                   }
                 }
+                }
+                
               }
-              
+
               if (x@type == "sparse") {
                 
-                tmp_info <- cocons::getDesignMatrix(model.list = x@model.list, data = x@data)
-                
-                theta_list <- cocons::getModelLists(theta = x@output$par, 
-                                                   par.pos = tmp_info$par.pos, type = "diff")
-                
-                X_std <- cocons::getScale(tmp_info$model.matrix,
-                                         mean.vector = x@info$mean.vector,
-                                         sd.vector = x@info$sd.vector
-                )
-                
                 spat_effects <- getSpatEffects(x)
                 
                 tmp_list <- list("x" = x@locs[, 1],
@@ -247,7 +237,7 @@ setMethod("plot",
                 tmp_list <- list("x" = x@locs[, 1],
                                  "y" = x@locs[, 2],
                                  "main" = "residuals",
-                                 "z" = x@z - getTrend(x))
+                                 "z" = x@z[,1] - getTrend(x))
                 
                 do.call(fields::quilt.plot, args = tmp_list)
                 
@@ -257,69 +247,83 @@ setMethod("plot",
                 fields::quilt.plot(x@locs, spat_effects$scale_x, ..., main = "approx. eff. scale")
                 fields::quilt.plot(x@locs, spat_effects$sd, ..., main = "se")
                 fields::quilt.plot(x@locs, spat_effects$nugget, ..., main = "nugget")
-              }
-              
-              if (!is.null(type)) {
-                if (x@type == "sparse" & type == "correlations") {
-                  tmp_info <- cocons::getDesignMatrix(
-                    model.list = x@model.list,
-                    data = x@data
-                  )
-                  
-                  theta_list <- cocons::getModelLists(
-                    theta = x@output$par,
-                    par.pos = tmp_info$par.pos, type = "diff"
-                  )
-                  
-                  X_std <- cocons::getScale(tmp_info$model.matrix)
-                  
-                  tmp_cov <- stats::cov2cor(cocons::cov_rns_smooth_taper_vector(
-                    theta_list, x@locs,
-                    X_std$std.covs
-                  )) ## !!!!!!!!!!!!!!!!!!!!!!!!!! FIXXXXXXXX
-                  
-                  for (ww in 1:length(index)) {
-                    fields::quilt.plot(x@locs, tmp_cov[index[ww], ])
+                
+                if (!is.null(type)) {
+                  if(type == "ellipse"){
+                    
+                    graphics::par(mfrow = c(1, 1))
+                    
+                    plot(x@locs, col = fields::tim.colors(128)[cut(x@z, 128)], pch = 20, cex = 1, asp = 1)
+                    
+                    number_x <- 10
+                    range_x <- range(x@locs[, 1])
+                    eps_x <- diff(range_x) / 3
+                    vals_x <- seq(from = range_x[1], to = range_x[2], length.out = number_x)
+                    
+                    for (ii in 1:length(vals_x)) {
+                      ref_y <- which(x@locs[, 1] < (vals_x[ii] + eps_x / 2) &
+                                       x@locs[, 1] > (vals_x[ii] - eps_x / 2))
+                      
+                      range_y <- range(x@locs[ref_y, 2])
+                      
+                      for (jj in 1:number_x) {
+                        center_locs <- c(vals_x[ii], range_y[1] + jj / number_x * (range_y[2] - range_y[1]))
+                        
+                        sss <- spam::nearest.dist(x = matrix(center_locs, ncol = 2), y = x@locs, delta = diff(range(x@locs))/5) 
+                        
+                        to_compute_sd <- sss@colindices[which.min(sss@entries)]
+                        
+                        .cocons.DrawEllipsoid(
+                          alpha_i = pi/2, 
+                          r = 1, 
+                          rho = spat_effects$scale_x[to_compute_sd],
+                          loc = x@locs[to_compute_sd, ], factr = factr
+                        )
+                      }
+                    }
                   }
                 }
+                
+                if (!is.null(type)) {
+                  if (type == "correlations") {
+                  }
+                }                
+                
               }
             }
 )
 
-#' Print Method for Coco Class
+#' Summary Method for Coco Class
 #'
-#' This method prints objects of class 'coco'.
-#' @name print
-#' @aliases print,coco-method
-#' @param x An object of class 'coco'.
-#' @param inv.hess inverse of the approximated hessian matrix (getHessian)
-#' @param ... Additional arguments to be passed to plot.
-#' @return print the coco object
+#' method summary for objects of class 'coco'.
+#' @name summary
+#' @aliases summary,coco-method
+#' @param object (\code{S4}) An object of class 'coco'.
+#' @param inv.hess (\code{numeric matrix} or \code{NULL}) inverse of the approximated hessian matrix (getHessian)
+#' @return summary the coco object
 #' @docType methods
-#' @exportMethod print
-#' @docType methods
-#' @rdname print-methods
+#' @rdname summary-methods
 #' @author Federico Blasi
 #' 
-setMethod("print", signature(x = "coco"), 
+setMethod("summary", signature(object = "coco"), 
           definition = 
-            function(x, inv.hess = NULL, ...){
+            function(object, inv.hess = NULL){
               
-              if(length(x@output) == 0){stop("object has not been fited yet.")}
+              if(length(object@output) == 0){stop("object has not been fited yet.")}
               
-              tmp_matrix <- cocons::getDesignMatrix(model.list = x@model.list, 
-                                                   data = x@data)
+              tmp_matrix <- cocons::getDesignMatrix(model.list = object@model.list, 
+                                                   data = object@data)
               
               adjusted_effects <- matrix(nrow = 7, ncol = dim(tmp_matrix$model.matrix)[2])
               colnames(adjusted_effects) <- colnames(tmp_matrix$model.matrix)
               
-              adjusted_eff_values <- cocons::getModelLists(x@output$par, 
+              adjusted_eff_values <- cocons::getModelLists(object@output$par, 
                                                           tmp_matrix$par.pos,
                                                           type = "diff") 
 
               if(!is.null(inv.hess)){
                 
-                Hess_mod <- getModHess(x, inv.hess = inv.hess)
+                Hess_mod <- getModHess(object, inv.hess = inv.hess)
                 
                 if(is.logical(tmp_matrix$par.pos$mean)){
                   number_mean <- sum(tmp_matrix$par.pos$mean)
@@ -368,7 +372,7 @@ setMethod("print", signature(x = "coco"),
                 
               }
               
-              if(x@type == "dense"){
+              if(object@type == "dense"){
                 
                 adjusted_effects[1, ] <- adjusted_eff_values$mean
                 adjusted_effects[2, ] <- adjusted_eff_values$std.dev
@@ -437,10 +441,10 @@ setMethod("print", signature(x = "coco"),
                 } else {
                   
                   cat(sprintf("%-15s %15s %15s %15s\n", "Output:", " ", " ", " "))
-                  cat(rep("-", 60), "\n")
+                  cat(rep("-", 65), "\n")
                   
                   cat(sprintf("%-15s %10s %10s %10s %10s %10s %10s %10s\n", "(raw)", "Mean", "Std. Dev.", "Scale", "Aniso.", "Tilt", "Smooth", "Nugget"))
-                  cat(rep("-", 60), "\n")
+                  cat(rep("-", 65), "\n")
                   
                   for (ii in 1:dim(adjusted_effects)[2]) {
                     cat(sprintf("%-15s %10s %10s %10s %10s %10s %10s %10s\n", 
@@ -476,9 +480,7 @@ setMethod("print", signature(x = "coco"),
                 
               }
               
-              if(x@type == "sparse"){
-                
-                # warning("taper estimations are biased.")
+              if(object@type == "sparse"){
                 
                 adjusted_effects[1, ] <- adjusted_eff_values$mean
                 adjusted_effects[2, ] <- adjusted_eff_values$std.dev
@@ -489,10 +491,10 @@ setMethod("print", signature(x = "coco"),
                 if(!is.null(inv.hess)){
                   
                   cat(sprintf("%-15s %15s %15s %15s\n", "Output:", " ", " ", " "))
-                  cat(rep("-", 65), "\n")
+                  cat(rep("-", 50), "\n")
                   
                   cat(sprintf("%-15s %15s %15s %15s %15s %15s\n", "(raw)", "Mean", "Std. Dev.", "Scale", "Smooth", "Nugget"))
-                  cat(rep("-", 65), "\n")
+                  cat(rep("-", 50), "\n")
                   
                   for (ii in 1:dim(adjusted_effects)[2]) {
                     cat(sprintf("%-15s %15s %15s %15s %15s %15s\n", 
@@ -534,10 +536,10 @@ setMethod("print", signature(x = "coco"),
                 } else {
                   
                   cat(sprintf("%-15s %15s %15s %15s\n", "Output:", " ", " ", " "))
-                  cat(rep("-", 60), "\n")
+                  cat(rep("-", 50), "\n")
                   
                   cat(sprintf("%-15s %10s %10s %10s %10s %10s\n", "(raw)", "Mean", "Std. Dev.", "Scale", "Smooth", "Nugget"))
-                  cat(rep("-", 60), "\n")
+                  cat(rep("-", 50), "\n")
                   
                   for (ii in 1:dim(adjusted_effects)[2]) {
                     cat(sprintf("%-15s %10s %10s %10s %10s %10s\n", 
@@ -563,7 +565,7 @@ setMethod("print", signature(x = "coco"),
                   
                 }
                 
-                cat(rep("-", 65), "\n")
+                cat(rep("-", 50), "\n")
                 
               
               }
@@ -575,11 +577,10 @@ setMethod("print", signature(x = "coco"),
 #' This method show objects of class 'coco'.
 #' @name show
 #' @aliases show,coco-method
-#' @param object An object of class 'coco'.
+#' @param object (\code{S4}) An object of class 'coco'.
 #' @return A plot is created.
 #' @docType methods
 #' @rdname show-methods
-#' @exportMethod show
 #' @author Federico Blasi
 #' 
 setMethod("show",
@@ -597,16 +598,12 @@ setMethod("show",
               
               # number of parameters
               
-              coco_info_light <- getDesignMatrix(model.list = object@model.list, data = object@data[1, , drop = FALSE])
-              
-              tmp_index <- lapply(coco_info_light$par.pos, FUN = is.logical)
-              
-              n_par <- sum(unlist(lapply(coco_info_light$par.pos, sum))[which(tmp_index == TRUE)])
-              
-              cat(sprintf("%-30s %30s\n", "model parameters", n_par))
+              cat(sprintf("%-30s %30s\n", "model parameters", .cocons.get.npars(object)))
               
               # covariates names
               
+              coco_info_light <- getDesignMatrix(model.list = object@model.list, data = object@data[1, , drop = FALSE])
+              
               cat(sprintf("%-30s %10s", "covariates ", paste0(colnames(coco_info_light$model.matrix),collapse = ", ")))
               
             }
@@ -621,17 +618,13 @@ setMethod("show",
               cat(sprintf("%-30s %30s\n", "dataset dim", paste0(dim(object@data), collapse = ", ")))
               
               # number of parameters
-              
-              coco_info_light <- getDesignMatrix(model.list = object@model.list,data = object@data[1, , drop = FALSE])
-              
-              tmp_index <- lapply(coco_info_light$par.pos, FUN = is.logical)
-              
-              n_par <- sum(unlist(lapply(coco_info_light$par.pos, sum))[which(tmp_index == TRUE)])
-              
-              cat(sprintf("%-30s %30s\n", "model parameters", n_par))
+
+              cat(sprintf("%-30s %30s\n", "model parameters", .cocons.get.npars(object)))
               
               # covariates names
               
+              coco_info_light <- getDesignMatrix(model.list = object@model.list, data = object@data[1, , drop = FALSE])
+              
               cat(sprintf("%-30s %10s", "covariates ", paste0(colnames(coco_info_light$model.matrix),collapse = ", ")))
               
             }
