@@ -1,18 +1,24 @@
 
-#' Optimizer of nonstationary spatial models
-#' @description Estimation of the spatial model parameters based on the L-BFGS-B optimizer \[1\]. 
-#' @details
-#' Current implementation only allows a single realization for \code{"pmle"} \code{optim.type}. 
-#' @usage cocoOptim(coco.object, boundaries = list(), 
-#' ncores = parallel::detectCores(), optim.control, optim.type)
-#' @param coco.object (\code{S4}) a \link{coco} object.
-#' @param boundaries (\code{list}) if provided, a list with lower, init, and upper values, as the one provided by \link{getBoundaries}. Otherwise,
-#' it is computed based on \link{getBoundaries} with global lower and upper values -2 and 2.
-#' @param ncores (\code{integer}) number of threads for the optimization routine.
-#' @param optim.control (\code{list}) list with settings to be passed to the optimParallel function \[2\].
-#' @param optim.type (\code{character}) Optimization approach: whether \code{"mle"} for classical Maximum Likelihood approach, 
-#' or \code{"pmle"} to factor out the spatial trend (when handling \code{"dense"} coco objects), or
-#' to factor out the global marginal standard deviation parameter (when considering \code{"sparse"} coco objects).
+#' Optimizer for Nonstationary Spatial Models
+#' 
+#' @description 
+#' This function estimates the spatial model parameters using the L-BFGS-B optimizer \[1\].
+#' 
+#' 
+#' @usage 
+#' cocoOptim(coco.object, boundaries = list(), ncores = "auto", 
+#' optim.type, safe, optim.control)
+#' 
+#' @param coco.object (\code{S4}) A \link{coco} object.
+#' @param boundaries (\code{list}) If provided, a list containing lower, initial, and upper values for the parameters, as defined by \link{getBoundaries}. If not provided, these values are automatically computed with global lower and upper bounds set to -2 and 2.
+#' @param ncores (\code{character} or \code{integer}) The number of threads to use for the optimization. If set to `"auto"`, the number of threads is chosen based on system capabilities or a fraction of the available cores.
+#' @param optim.type (\code{character}) The optimization approach. Options include:
+#' \itemize{
+#'   \item \code{"mle"}: Classical Maximum Likelihood estimation.
+#'   \item \code{"pmle"}: Profile Maximum Likelihood, factoring out the spatial trend for dense objects or the global marginal variance parameter for sparse objects.
+#' }
+#' @param safe (\code{logical}) If `TRUE`, the function avoids Cholesky decomposition errors due to ill-posed covariance matrices by returning a pre-defined large value. Defaults to `TRUE`.
+#' @param optim.control (\code{list}) A list of settings to be passed to the \link[optimParallel]{optimParallel} function \[2\].
 #' @returns (\code{S4}) An optimized S4 object of class \code{coco}.
 #' @author Federico Blasi
 #' @seealso [\link[optimParallel]{optimParallel}]
@@ -42,27 +48,42 @@
 #' boundaries = getBoundaries(coco_object,
 #' lower.value = -3, 3))
 #' 
+#' plotOptimInfo(optim_coco)
+#' 
 #' plot(optim_coco)
 #' 
-#' print(optim_coco)
+#' plot(optim_coco, type = 'ellipse')
 #' 
+#' plot(optim_coco, type = 'correlations', index = c(2,3,5))
+#' 
+#' summary(optim_coco)
+#'  
 #' getEstims(optim_coco)
+#' 
 #' }
 #' 
 cocoOptim <- function(coco.object, boundaries = list(), 
-                       ncores = parallel::detectCores(), 
-                       optim.control = NULL, optim.type = "mle"){
-  
-  .cocons.check.ncores(ncores)
+                      ncores = "auto", optim.type = "mle",
+                      safe = TRUE,
+                      optim.control = NULL){
   
   # Init objects
   if(T){
     
+    if(is.character(ncores)){
+      ncores <- .cocons.set.ncores(coco.object, optim.control)
+    } else{
+      .cocons.check.ncores(ncores)
+    }
+    
+    cat("using ",ncores," threads.\n")
+    
     designMatrix <- cocons::getDesignMatrix(
       model.list = coco.object@model.list,
       data = coco.object@data
     )
     
+    # Check boundaries
     if(length(boundaries) == 0){
       boundaries <- cocons::getBoundaries(
         x = coco.object, 
@@ -72,30 +93,39 @@ cocoOptim <- function(coco.object, boundaries = list(),
       .cocons.check.boundaries(coco.object, boundaries)
     }
     
-    # Categorical variables
-    # problem here: i can specify categorical values in data but then not use any in formula
-    if(!is.null(coco.object@info$cat.vars)){
+    # Skip scaling
+    if(any(colnames(coco.object@data[,coco.object@info$skip.scale]) %in% 
+           colnames(coco.object@data))){
       tmp_info <- .cocons.setDesignMatrixCat(coco.object, designMatrix = designMatrix)
       tmp_values <- tmp_info$tmp_values
-      empty_matrix <- tmp_info$empty_matrix
-    } else{
+      mod_DM <- tmp_info$mod_DM
+    } else {
       tmp_values <- cocons::getScale(designMatrix$model.matrix)
-      empty_matrix <- tmp_values$std.covs
+      mod_DM <- tmp_values$std.covs      
     }
     
+    # if optim.control not provided, then some general Optim.control is provided
+    optim.control <- if (is.null(optim.control)) {
+      getOption("cocons.Optim.Control")
+    } else {
+      .cocons.update.optim.control(optim.control)
+    }
+    
+    # Suggested by OptimParallel
+    if(tolower(.Platform$OS.type) != "windows"){
+      cl <- parallel::makeCluster(spec = ncores, type = "FORK", outfile = "")
+    } else{
+      cl <- parallel::makeCluster(spec = ncores, outfile = "")
+    }
+    
+    on.exit(parallel::stopCluster(cl), add = TRUE)
+    
   }
   
   if (coco.object@type == "dense") {
     
     if(optim.type == "mle"){
-      
-      # Suggested by OptimParallel
-      if(tolower(.Platform$OS.type) != "windows"){
-        cl <- parallel::makeCluster(spec = ncores, type = "FORK", outfile = "")
-      } else{
-        cl <- parallel::makeCluster(spec = ncores, outfile = "")
-      }
-      
+
       parallel::setDefaultCluster(cl = cl)
       parallel::clusterEvalQ(cl, library("cocons"))
       
@@ -108,29 +138,21 @@ cocoOptim <- function(coco.object, boundaries = list(),
         "n" = dim(coco.object@z)[1],
         "smooth.limits" = coco.object@info$smooth.limits,
         "z" = coco.object@z,
-        "x_covariates" = empty_matrix,
+        "x_covariates" = mod_DM,
         "par.pos" = designMatrix$par.pos,
         "lambda" = coco.object@info$lambda,
-        "locs" = coco.object@locs
+        "locs" = coco.object@locs,
+        "safe" = safe
       )
       
       rm(designMatrix)
-      
-      # if optim.control not provided, then some general Optim.control is provided
-      if (is.null(optim.control)) {
-        optim.control <- getOption("cocons.Optim.Control")
-      } else{
-        optim.control <- .cocons.update.optim.control(optim.control)
-      }
-      
+
       # Call optim routine
       output_dense <- do.call(what = optimParallel::optimParallel, args = c(
         args_optim,
         optim.control
       ))
       
-      parallel::stopCluster(cl)
-      
       .cocons.check.convergence(output_dense, boundaries)
       
       coco.object@output <- output_dense
@@ -138,6 +160,8 @@ cocoOptim <- function(coco.object, boundaries = list(),
       coco.object@info$mean.vector <- tmp_values$mean.vector
       coco.object@info$sd.vector <- tmp_values$sd.vector
       coco.object@info$optim.type <- "mle"
+      coco.object@info$safe <- safe
+      coco.object@info$call <- match.call()
       
       return(coco.object)
       
@@ -145,72 +169,49 @@ cocoOptim <- function(coco.object, boundaries = list(),
     
     if(optim.type == "pmle"){
       
-      if(dim(coco.object@z)[2] > 1){stop('profile ML routines only handle single realizations.')}
-      
       if(!is.logical(designMatrix$par.pos$mean)){stop("profile ML only available when considering covariates in the mean.")}
 
-      if(is.logical(designMatrix$par.pos$mean)){
-        x_betas <- empty_matrix[, designMatrix$par.pos$mean]
-      }
+      x_betas <- mod_DM[, designMatrix$par.pos$mean]
       
       tmp_boundaries <- boundaries
       
-      # profiling both betas
+      # profiling betas
       boundaries$theta_init <- boundaries$theta_init[-c(1:(sum(designMatrix$par.pos$mean)))]
       boundaries$theta_upper <- boundaries$theta_upper[-c(1:(sum(designMatrix$par.pos$mean)))]
       boundaries$theta_lower <- boundaries$theta_lower[-c(1:(sum(designMatrix$par.pos$mean)))]
       
-      # Suggested by OptimParallel
-      if(tolower(.Platform$OS.type) != "windows"){
-        cl <- parallel::makeCluster(spec = ncores, type = "FORK", outfile = "")
-      } else{
-        cl <- parallel::makeCluster(spec = ncores, outfile = "")
-      }
-      
       parallel::setDefaultCluster(cl = cl)
       parallel::clusterEvalQ(cl, library("cocons"))
       
       # factoring out betas
-      tmp_par_pos <- designMatrix$par.pos
-      
-      if(is.logical(tmp_par_pos$mean)){
-        # factoring out all betas
-        tmp_par_pos$mean <- rep(FALSE, length(tmp_par_pos$mean))
+      if(T){
+        tmp_par_pos <- designMatrix$par.pos
+        tmp_par_pos$mean <- rep(FALSE, length(tmp_par_pos$mean))    
       }
-      
+
       args_optim <- list(
         "fn" = cocons::GetNeg2loglikelihoodProfile,
         "method" = "L-BFGS-B",
         "lower" = boundaries$theta_lower,
         "par" = boundaries$theta_init,
         "upper" = boundaries$theta_upper,
-        "n" = length(coco.object@z),
+        "n" = dim(coco.object@z)[1],
         "smooth.limits" = coco.object@info$smooth.limits,
         "z" = coco.object@z,
-        "x_covariates" = empty_matrix,
+        "x_covariates" = mod_DM,
         "par.pos" = tmp_par_pos,
         "lambda" = coco.object@info$lambda,
         "locs" = coco.object@locs,
-        "x_betas" = x_betas
+        "x_betas" = x_betas,
+        "safe" = safe
       )
       
-      # rm(designMatrix)
-      
-      # if optim.control not provided, then some general Optim.control is provided
-      if (is.null(optim.control)) {
-        optim.control <- getOption("cocons.Optim.Control")
-      } else{
-        optim.control <- .cocons.update.optim.control(optim.control)
-      }
-      
       # Call optim routine
       output_dense <- do.call(what = optimParallel::optimParallel, args = c(
         args_optim,
         optim.control
       ))
       
-      parallel::stopCluster(cl)
-      
       theta_list <- cocons::getModelLists(theta = output_dense$par, 
                                          par.pos = args_optim$par.pos, type = "diff")
       
@@ -218,10 +219,23 @@ cocoOptim <- function(coco.object, boundaries = list(),
                                   x_covariates  =  args_optim$x_covariates,
                                   smooth_limits = args_optim$smooth.limits)
       
-      Sigma_X <- solve(Sigma_cpp, args_optim$x_betas)
-      betass <- solve(t(args_optim$x_betas) %*% Sigma_X) %*% t(Sigma_X) %*% args_optim$z
+      # Compute Betas
+
+      if(T){
+        L <- chol(Sigma_cpp)
+        V <- backsolve(L, forwardsolve(L, x_betas,
+                                       transpose = TRUE, 
+                                       upper.tri = TRUE))
+        W <- crossprod(x_betas, V)
+        betass <- c(solve(W, t(V)) %*% rowSums(args_optim$z)) / dim(args_optim$z)[2]
+        names(betass) <- colnames(x_betas)
+      }
+      
       output_dense$par <- c(betass, output_dense$par)
       
+      tmp_boundaries$theta_upper[1:(sum(designMatrix$par.pos$mean))] <- rep(Inf,sum(designMatrix$par.pos$mean))
+      tmp_boundaries$theta_lower[1:(sum(designMatrix$par.pos$mean))] <- rep(-Inf,sum(designMatrix$par.pos$mean))
+      
       .cocons.check.convergence(output_dense, tmp_boundaries)
       
       coco.object@output <- output_dense
@@ -229,6 +243,8 @@ cocoOptim <- function(coco.object, boundaries = list(),
       coco.object@info$mean.vector <- tmp_values$mean.vector
       coco.object@info$sd.vector <- tmp_values$sd.vector
       coco.object@info$optim.type <- "pmle"
+      coco.object@info$safe <- safe
+      coco.object@info$call <- match.call()
       
       return(coco.object)
       
@@ -247,20 +263,11 @@ cocoOptim <- function(coco.object, boundaries = list(),
       )
       
       print(summary(ref_taper))
-      
-      # Suggested by OptimParallel
-      if(tolower(.Platform$OS.type) != "windows"){
-        cl <- parallel::makeCluster(spec = ncores, type = "FORK", outfile = "",
-                                    useXDR = FALSE)
-      } else{
-        cl <- parallel::makeCluster(spec = ncores, outfile = "", 
-                                    useXDR = FALSE)
-      }
-      
+
       parallel::setDefaultCluster(cl = cl)
       parallel::clusterEvalQ(cl, library("cocons"))
       parallel::clusterEvalQ(cl, library("spam"))
-      parallel::clusterEvalQ(cl, library("spam64"))
+      #parallel::clusterEvalQ(cl, library("spam64"))
       parallel::clusterEvalQ(cl, options(spam.cholupdatesingular = "error"))
       parallel::clusterEvalQ(cl, options(spam.cholsymmetrycheck = FALSE))
       
@@ -273,32 +280,21 @@ cocoOptim <- function(coco.object, boundaries = list(),
         "par.pos" = designMatrix$par.pos,
         "ref_taper" = ref_taper,
         "locs" = coco.object@locs,
-        "x_covariates" = empty_matrix,
+        "x_covariates" = mod_DM,
         "smooth.limits" = coco.object@info$smooth.limits,
         "cholS" = spam::chol.spam(ref_taper),
         "z" = coco.object@z,
-        "n" = length(coco.object@z),
-        "lambda" = coco.object@info$lambda
+        "n" = dim(coco.object@z)[1],
+        "lambda" = coco.object@info$lambda,
+        "safe" = safe
       )
       
-      #rm(designMatrix)
-      #rm(ref_taper)
-      
-      # if optim.control not provided, then some general Optim.control is provided
-      if (is.null(optim.control)) {
-        optim.control <- getOption("cocons.Optim.Control")
-      } else{
-        optim.control <- .cocons.update.optim.control(optim.control)
-      }
-      
       # Call optim routine
       output_taper <- do.call(what = optimParallel::optimParallel, args = c(
         args_optim,
         optim.control
       ))
       
-      parallel::stopCluster(cl)
-      
       .cocons.check.convergence(output_taper, boundaries)
       
       coco.object@output <- output_taper
@@ -306,13 +302,15 @@ cocoOptim <- function(coco.object, boundaries = list(),
       coco.object@info$mean.vector <- tmp_values$mean.vector
       coco.object@info$sd.vector <- tmp_values$sd.vector
       coco.object@info$optim.type <- "mle"
-
+      coco.object@info$safe <- safe
+      coco.object@info$call <- match.call()
+      
       return(coco.object)
     }
     
     if(optim.type == "pmle"){
       
-      if(dim(coco.object@z)[2] > 1){stop('profile ML routines only handle single realizations.')}
+      if(!is.logical(designMatrix$par.pos$std.dev)){stop("at least a global sigma needs to be estimated for sparse pmle coco objects.")}
       
       # taper
       ref_taper <- coco.object@info$taper(
@@ -321,34 +319,23 @@ cocoOptim <- function(coco.object, boundaries = list(),
       )
       
       print(summary(ref_taper))
-      
-      # Suggested by OptimParallel
-      if(tolower(.Platform$OS.type) != "windows"){
-        cl <- parallel::makeCluster(spec = ncores, type = "FORK", outfile = "",
-                                    useXDR = FALSE)
-      } else{
-        cl <- parallel::makeCluster(spec = ncores, outfile = "", 
-                                    useXDR = FALSE)
-      }
-      
+
       parallel::setDefaultCluster(cl = cl)
       parallel::clusterEvalQ(cl, library("cocons"))
       parallel::clusterEvalQ(cl, library("spam"))
-      parallel::clusterEvalQ(cl, library("spam64"))
+      #parallel::clusterEvalQ(cl, library("spam64"))
       parallel::clusterEvalQ(cl, options(spam.cholupdatesingular = "error"))
       parallel::clusterEvalQ(cl, options(spam.cholsymmetrycheck = FALSE))
       
-      # options(spam.cholsymmetrycheck = FALSE)
-      
       tmp_par.pos <- designMatrix$par.pos
       
-      if(is.logical(tmp_par.pos$std.dev)){
-        tmp_par.pos$std.dev[1] <- FALSE
-      }
-      
+      tmp_par.pos$std.dev[1] <- FALSE
+
       boundaries_temp <- boundaries
       
-      first_sigma <- sum(designMatrix$par.pos$mean) + 1
+      first_sigma <- if(is.logical(designMatrix$par.pos$mean)){
+        first_sigma <- sum(designMatrix$par.pos$mean) + 1
+      } else{ first_sigma <- 1}
       
       boundaries$theta_init <- boundaries$theta_init[-first_sigma]
       boundaries$theta_upper <- boundaries$theta_upper[-first_sigma]
@@ -363,32 +350,21 @@ cocoOptim <- function(coco.object, boundaries = list(),
         "par.pos" = tmp_par.pos,
         "ref_taper" = ref_taper,
         "locs" = coco.object@locs,
-        "x_covariates" = empty_matrix,
+        "x_covariates" = mod_DM,
         "smooth.limits" = coco.object@info$smooth.limits,
         "cholS" = spam::chol.spam(ref_taper),
         "z" = coco.object@z,
-        "n" = length(coco.object@z),
-        "lambda" = coco.object@info$lambda
+        "n" = dim(coco.object@z)[1],
+        "lambda" = coco.object@info$lambda,
+        "safe" = safe
       )
       
-      #rm(designMatrix)
-      #rm(ref_taper)
-      
-      # if optim.control not provided, then some general Optim.control is provided
-      if (is.null(optim.control)) {
-        optim.control <- getOption("cocons.Optim.Control")
-      } else{
-        optim.control <- .cocons.update.optim.control(optim.control)
-      }
-      
       # Call optim routine
       output_taper <- do.call(what = optimParallel::optimParallel, args = c(
         args_optim,
         optim.control
       ))
       
-      parallel::stopCluster(cl)
-      
       if(T){
         
         theta_list <- getModelLists(theta = output_taper$par, par.pos = args_optim$par.pos, type = "diff")
@@ -400,21 +376,33 @@ cocoOptim <- function(coco.object, boundaries = list(),
                                                                                                      rowpointers = args_optim$ref_taper@rowpointers,
                                                                                                      smooth_limits =  args_optim$smooth.limits)
         
-        resid <- c(args_optim$z - args_optim$x_covariates %*% theta_list$mean)
+        cum_qprod <- sum(apply(args_optim$z, MARGIN = 2, FUN = function(x){
+          resid <- c(x - args_optim$x_covariates %*% theta_list$mean)
+          SigmaX <- spam::solve(args_optim$ref_taper, resid)
+          c(resid %*% SigmaX)
+        }))
         
-        SigmaX <- spam::solve(args_optim$ref_taper, resid)
-        sigma_0 <- c(resid %*% SigmaX) / args_optim$n
+        sigma_0 <- cum_qprod / (args_optim$n * dim(args_optim$z)[2])
         
       }
       
-      first_scale <- sum(args_optim$par.pos$mean) + sum(args_optim$par.pos$std.dev) + 1
+      first_scale <- if(is.logical(designMatrix$par.pos$mean)){
+        sum(args_optim$par.pos$mean) + sum(args_optim$par.pos$std.dev) + 1
+      } else{
+        sum(args_optim$par.pos$std.dev) + 1
+      }
       
-      tmp_global_scale <- output_taper$par[first_scale]
+      # what if scale is fixed (?) -- implemented v0.1
+        
+      tmp_global_scale <- if(is.logical(args_optim$par.pos$scale)){output_taper$par[first_scale]} else{
+        tmp_global_scale <- args_optim$par.pos$scale[1]
+      }
       
       first_par <- log(sigma_0) + tmp_global_scale
       names(first_par) <- "std.dev.limits"
       second_par <- log(sigma_0) - tmp_global_scale
-      
+      names(second_par) <- "scale.limits"
+        
       if(is.logical(args_optim$par.pos$mean)){
         new_pars <- output_taper$par[1:(sum(args_optim$par.pos$mean))]
       } else{new_pars <- numeric(0)}
@@ -427,24 +415,33 @@ cocoOptim <- function(coco.object, boundaries = list(),
       
       if(is.logical(args_optim$par.pos$scale)){
         new_pars <- c(new_pars, second_par, output_taper$par[(first_scale+1):(first_scale + sum(args_optim$par.pos$scale) - 1)])
+        first_smooth <- first_scale + if(is.logical(args_optim$par.pos$scale)){sum(args_optim$par.pos$scale)}
+      } else{
+        first_smooth <- first_scale
       }
       
-      first_smooth <- first_scale + sum(args_optim$par.pos$scale)
-      
       if(is.logical(args_optim$par.pos$smooth)){
         new_pars <- c(new_pars, output_taper$par[first_smooth:(first_smooth + sum(args_optim$par.pos$smooth) - 1)])
-      }
+        first_nugget <- first_smooth + if(is.logical(args_optim$par.pos$smooth)){sum(args_optim$par.pos$smooth)}
+      } else{first_nugget <- first_smooth}
       
-      first_nugget <- first_smooth + sum(args_optim$par.pos$smooth)
-      
-      if(sum(args_optim$par.pos$nugget) > 0){
+      if(is.logical(args_optim$par.pos$nugget)){
         new_pars <- c(new_pars, output_taper$par[first_nugget:(first_nugget + sum(args_optim$par.pos$nugget) - 1)])
       }
       
       output_taper$par <- new_pars
       
       # fix names output_taper
-
+      
+      boundaries_temp$theta_lower[first_sigma] <- -Inf
+      boundaries_temp$theta_upper[first_sigma] <- Inf
+      
+      if(is.logical(args_optim$par.pos$scale)){
+        tmp_value <- boundaries_temp$theta_upper[first_scale]
+        boundaries_temp$theta_upper[first_scale + 1] <- log(sigma_0) - boundaries_temp$theta_lower[first_scale]
+        boundaries_temp$theta_lower[first_scale + 1] <- log(sigma_0) - tmp_value        
+      }
+      
       .cocons.check.convergence(output_taper, boundaries_temp)
       
       coco.object@output <- output_taper
@@ -452,7 +449,9 @@ cocoOptim <- function(coco.object, boundaries = list(),
       coco.object@info$mean.vector <- tmp_values$mean.vector
       coco.object@info$sd.vector <- tmp_values$sd.vector
       coco.object@info$optim.type <- "pmle"
-
+      coco.object@info$safe <- safe
+      coco.object@info$call <- match.call()
+      
       return(coco.object)
     }
     
